"""Takeoff Object SDK"""

# ───────────────────────────────────────────────────── imports ────────────────────────────────────────────────────── #

import os
import time
import yaml

from loguru import logger

from .schema import TakeoffEnvSetting
from .utils import is_docker_logs_error, is_takeoff_loading, start_takeoff, start_takeoff_with_manifest

# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #
#                                                       Takeoff                                                        #
# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #


class Takeoff:
    def __init__(self, model_name=None, device=None, use_manifest=False, **kwargs):
        self.model_name = model_name
        self.device = device

        self.models = []
        self.consumer_groups = {}
        self.server_url = None
        self.management_url = None
        self.openai_url = None
        self.container = None

        self.use_manifest = use_manifest

        if not self.use_manifest:
            self.takeoff_config = TakeoffEnvSetting(model_name=model_name, device=device, **kwargs)
            self.num_readers = 1

    def cleanup(self):
        """Remove the Takeoff container if it exists"""
        if self.container:
            self.container.remove(force=True)

    @classmethod
    def from_config(cls, takeoff_config: TakeoffEnvSetting):
        """Create a Takeoff object from a TakeoffEnvSetting object

        Args:
            takeoff_config (TakeoffEnvSetting): Takeoff configuration

        Returns:
            Takeoff: Takeoff object
        """

        takeoff_args = takeoff_config.model_dump()
        return cls(**takeoff_args)

    @classmethod
    def from_manifest(cls, manifest_path: str, device: str = "cuda"):
        """Create a Takeoff object from a manifest file

        Args:
            manifest_path (str): path to the manifest file
            device (str, optional): device choice. Defaults to "cuda".

        Raises:
            FileNotFoundError: if the manifest file is not found

        Returns:
            Takeoff: Takeoff object
        """

        # turns the user input manifest path into an absolute path
        manifest_path = os.path.abspath(manifest_path)

        if not os.path.exists(manifest_path):
            raise FileNotFoundError(f"Manifest file not found at {manifest_path}")

        yaml_file = yaml.safe_load(open(manifest_path, "r"))

        cls.num_readers = len(yaml_file["takeoff"]["readers_config"].keys())
        cls.manifest_path = manifest_path

        return cls(device=device, use_manifest=True)

    def start(self):
        """Start the Takeoff server and add the model to the primary consumer group"""

        logger.info("Starting Takeoff server...")
        if self.use_manifest:
            logger.info(f"Using manifest file: {self.manifest_path}")
            (
                self.takeoff_port,
                self.management_port,
                self.openai_port,
                self.prometheus_port,
                self.container,
            ) = start_takeoff_with_manifest(self.manifest_path, self.device)
        else:
            (
                self.takeoff_port,
                self.management_port,
                self.openai_port,
                self.prometheus_port,
                self.container,
            ) = start_takeoff(self.takeoff_config)

        self.server_url = f"http://localhost:{self.takeoff_port}"
        self.management_url = f"http://localhost:{self.management_port}"
        self.openai_url = f"http://localhost:{self.openai_port}"
        self.prometheus_url = f"http://localhost:{self.prometheus_port}"

        logger.info(f"Takeoff server running at: {self.server_url}")
        logger.info(f"Takeoff management server running at: {self.management_url}")
        logger.info(f"Takeoff openai compatible server running at: {self.openai_url}")
        logger.info(f"Takeoff prometheus server running at: {self.prometheus_url}")

        for _ in range(200):  # Some models take a while to download
            if not is_takeoff_loading(self.server_url, self.num_readers):
                break

            traceback = is_docker_logs_error(self.container)

            if traceback is not None:
                logger.error("Takeoff server failed to start. Error in docker logs. Cleaning up...")
                self.cleanup()
                raise Exception(
                    "Takeoff server failed to start due to error in docker logs. See below for details. \n" + traceback
                )
            logger.info("building...")
            time.sleep(3)
        else:
            raise Exception("Takeoff server build timed out")

        logger.info("server ready!")



"""Utility functions for the Takeoff Cli."""

from pathlib import Path
import platform
import re

import psutil
from InquirerPy import inquirer
from InquirerPy.validator import EmptyInputValidator


def check_docker_environment():
    """Check if Docker is installed and running."""

    try:
        import docker

        client = docker.from_env()
        client.containers.list()
        return True
    except Exception:
        return False


def check_device_environment():
    """Check system device information."""

    return get_system_info()


def check_license_key_format(license_key: str) -> bool:
    """Check if the license key format is valid.

    Args:
        license_key (str): Input license key.

    Returns:
        bool: True if the license key format is valid, False otherwise.
    """
    pattern = r"^[A-Z0-9]{5}-[A-Z0-9]{5}-[A-Z0-9]{5}-[A-Z0-9]{5}$"
    return bool(re.match(pattern, license_key))


def bytes_to_gb(bytes: int) -> float:
    """Convert bytes to gigabytes."""
    return bytes / (1024**3)


def get_system_info() -> dict:
    try:
        import pynvml

        pynvml.nvmlInit()
        gpu_info = {}
        device_count = pynvml.nvmlDeviceGetCount()
        for i in range(device_count):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            memory_info = pynvml.nvmlDeviceGetMemoryInfo(handle)

            gpu_info[f"GPU {i}"] = {
                "name": pynvml.nvmlDeviceGetName(handle),
                "memory": round(bytes_to_gb(memory_info.total), 2),
                "memory_unit": "GB",
                "driver": pynvml.nvmlSystemGetDriverVersion(),
            }
        pynvml.nvmlShutdown()
    except ImportError:
        gpu_info = "NVIDIA library (pynvml) not installed or no NVIDIA GPU detected."  # type: ignore

    cpu_info = {
        "processor": platform.processor(),
        "architecture": platform.machine(),
        "physical_cores": psutil.cpu_count(logical=False),
        "total_cores": psutil.cpu_count(logical=True),
    }

    return {"CPU Info": cpu_info, "GPU Info": gpu_info}


def pull_image(image_name: str, tag: str = "latest") -> None:
    import docker

    client = docker.from_env()
    try:
        client.images.pull(image_name, tag=tag)
    except docker.errors.APIError as e:
        raise Exception(f"Error pulling image {image_name}:{tag}. {e}")


def get_model_from_user(models):
    """
    Show a selection list to the user with the models available.

    :param models: A list of model names (strings) available for selection.
    :return: The name of the model selected by the user.
    """
    model_name = inquirer.select(
        message="We provide a list of models to choose from. If you have a custom model, please choose 'Others'.",
        choices=models,
        validate=EmptyInputValidator("A selection is required."),
        vi_mode=False,
        cycle=True,
    ).execute()

    return model_name


def run_license_validator(license_key) -> dict:
    """Run the license validator in a Docker container and return the results.

    Args:
        license_key (str): The license key to validate.

    Returns:
        dict: The results of the license validation, including the exit code and output.
    """
    import docker

    # Initialize a Docker client
    client = docker.from_env()

    volumes = [f"{Path.home()}/.takeoff_cache:/code/models"]

    # Run the container with the license key as a command-line argument
    container = client.containers.run(
        "tytn/license-validation:latest", command=["22918", license_key], remove=False, detach=True, volumes=volumes
    )

    # Wait for the container to finish running
    result = container.wait()

    # Retrieve the exit code and output
    exit_code = result["StatusCode"]
    output = container.logs().decode("utf-8")

    container.remove()

    # Return the results
    return {"exit_code": exit_code, "output": output}



"""This script is used to bump the version of the package. pyproject.toml is used to store the version."""
import re
import sys
import argparse

import toml

import requests


def check_version_on_pypi(package_name: str, version: str) -> bool:
    """Checks if the given version of the package exists on PyPI.

    Args:
        package_name (str): The name of the package
        version (str): The version to check

    Returns:
        bool: True if the version exists, False otherwise
    """
    url = f"https://pypi.org/pypi/{package_name}/json"
    try:
        response = requests.get(url)
        response.raise_for_status()  # Raise an exception for HTTP errors
        data = response.json()
        versions = data.get("releases", {}).keys()
        return version in versions
    except requests.RequestException as e:
        print(f"Error checking version on PyPI: {e}")
        return False


def read_version():
    with open("pyproject.toml", "r") as file:
        data = toml.load(file)
    return data["project"]["version"]


def read_package_name():
    with open("pyproject.toml", "r") as file:
        data = toml.load(file)
    return data["project"]["name"]


def write_version(new_version):
    # Read the content of the file
    with open("pyproject.toml", "r") as file:
        content = file.readlines()

    # Find and update the version line
    for i, line in enumerate(content):
        if line.startswith("version = "):
            content[i] = f'version = "{new_version}"\n'
            break

    # Write the updated content back to the file
    with open("pyproject.toml", "w") as file:
        file.writelines(content)


if __name__ == "__main__":
    # Parse the command-line arguments
    parser = argparse.ArgumentParser(description="Bump the package version.")
    parser.add_argument("--version", required=True, help="The new version number to apply.")
    args = parser.parse_args()

    new_version = args.version

    package_name = read_package_name()

    version_pattern = r"^\d+\.\d+\.\d+$"

    # check if the new_version is provided
    if not new_version:
        print("Usage: python3 bump_version.py x.x.x")
        sys.exit(1)

    # Check if the new_version matches the required pattern
    if not re.match(version_pattern, new_version):
        print("Error: Version must be in the format x.x.x, where x is a digit.")
        sys.exit(1)

    # Check if the new version already exists on PyPI

    if check_version_on_pypi(package_name, new_version):
        print(f"Error: Version {new_version} already exists on PyPI.")
        sys.exit(1)

    # if all checks pass, update the version
    write_version(new_version)
    print(f"Version updated to: {new_version}")



from unittest.mock import MagicMock, patch

import pytest

from takeoff_launcher import Takeoff, TakeoffEnvSetting


def test_takeoff_initialization():
    model_name = "test_model"
    device = "cpu"
    takeoff = Takeoff(model_name=model_name, device=device)

    assert takeoff.model_name == model_name
    assert takeoff.device == device
    assert isinstance(takeoff.takeoff_config, TakeoffEnvSetting)
    assert takeoff.models == []
    assert takeoff.consumer_groups == {}
    assert takeoff.server_url is None
    assert takeoff.management_url is None
    assert takeoff.container is None


def test_takeoff_from_config():
    config = TakeoffEnvSetting(model_name="test_model", device="cpu")
    takeoff = Takeoff.from_config(config)

    assert takeoff.model_name == config.model_name
    assert takeoff.device == config.device


def test_takeoff_from_manifest():
    manifest_path = "tests/test_manifest.yaml"
    takeoff = Takeoff.from_manifest(manifest_path)

    assert takeoff.model_name is None
    assert takeoff.device == "cuda"

    assert takeoff.use_manifest is True


@patch("takeoff_launcher.sdk.takeoff.is_docker_logs_error", return_value=None)
@patch("takeoff_launcher.sdk.takeoff.start_takeoff", return_value=(8000, 8001, 8003, MagicMock()))
@patch("takeoff_launcher.sdk.takeoff.is_takeoff_loading", side_effect=[True, True, False])
def test_takeoff_start(mock_is_loading, mock_start_takeoff, mock_logs_error):
    takeoff = Takeoff(model_name="test_model", device="cpu")

    # Assuming readers() method returns some mock data
    mock_readers = {"group1": [{"model_name": "model1"}]}
    takeoff.readers = MagicMock(return_value=mock_readers)

    takeoff.start()

    assert takeoff.server_url == "http://localhost:8000"
    assert takeoff.management_url == "http://localhost:8001"
    mock_start_takeoff.assert_called_once_with(takeoff.takeoff_config)

    assert mock_is_loading.is_called


@patch("takeoff_launcher.sdk.takeoff.is_docker_logs_error", return_value="Error in logs")
@patch("takeoff_launcher.sdk.takeoff.start_takeoff", return_value=(8000, 8001, 8003, MagicMock()))
@patch("takeoff_launcher.sdk.takeoff.is_takeoff_loading", side_effect=[True, True, False])
def test_takeoff_start_error_in_docker_logs(mock_is_loading, mock_start_takeoff, mock_logs_error):
    takeoff = Takeoff(model_name="test_model", device="cpu")

    # Assuming readers() method returns some mock data
    mock_readers = {"group1": [{"model_name": "model1"}]}
    takeoff.readers = MagicMock(return_value=mock_readers)

    with pytest.raises(Exception) as e:
        takeoff.start()
    assert "Takeoff server failed to start due to error in docker logs. See below for details. \nError in logs" in str(
        e.value
    )



"""Takeoff CLI for starting and managing Takeoff servers."""
# ───────────────────────────────────────────────────── imports ────────────────────────────────────────────────────── #

import os

import typer
from loguru import logger
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Prompt
from rich.table import Table

from .sdk.takeoff import Takeoff
from .utils import (
    check_docker_environment,
    check_license_key_format,
    get_model_from_user,
    get_system_info,
    pull_image,
    run_license_validator,
)


# ───────────────────────────────────────────────── Global Settings ────────────────────────────────────────────────── #
main = typer.Typer(no_args_is_help=True)
console = Console()

# remove default logger to avoid logs in takeoff-sdk
logger.remove()


DEFAULT_MODELS = [
    "HuggingFaceH4/zephyr-7b-gemma-v0.1",
    "HuggingFaceH4/zephyr-7b-beta",
    "NousResearch/Llama-2-7b-chat-hf",
    "TitanML/llama2-7b-base-4bit-AWQ",
    "TheBloke/CodeLlama-7B-AWQ",
    "TheBloke/Mistral-7B-Instruct-v0.2-AWQ",
    "Others",
]

# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #
#                                                Takeoff Start Command                                                 #
# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #


@main.command()
def start():
    """Quickly start command. Type `takeoff start` to prompt the user for a quick start."""

    # ─────────────────────────────────── check docker environment ─────────────────────────────────── #
    docker_installed = check_docker_environment()
    if not docker_installed:
        console.print(
            "[bold red]:x: Docker environment check failed. Please make sure Docker is installed and running.[/]"
        )
        return
    console.print("[bold green]:heavy_check_mark: Docker environment check passed[/]")

    # Create a Docker client
    import docker

    client = docker.from_env()

    # ─────────────────────────────────── check system environment ─────────────────────────────────── #
    sys_info = get_system_info()
    _, gpu_info = sys_info["CPU Info"], sys_info["GPU Info"]

    if isinstance(gpu_info, str):
        has_gpu = False
        # no gpu detected, we warn the user
        console.print(
            "[bold yellow]:warning: No Nvidia GPU detected. We recommend using a GPU for optimal performance.[/]"
        )
    else:
        has_gpu = True
        # gpu detected, we can print the gpu name and memory
        gpu_name = gpu_info["GPU 0"]["name"]
        gpu_memory = gpu_info["GPU 0"]["memory"]
        console.print(
            f"[bold green]:heavy_check_mark: Nvidia GPU detected:[/] [bold blue]{gpu_name}[/] with [bold blue]{gpu_memory}GB[/] memory."
        )

    # ────────────────────────────────────── check license key ─────────────────────────────────────── #

    # check if the user has a license key
    license_res_cache = run_license_validator("TEST")
    exit_code = license_res_cache["exit_code"]

    license_key = ""

    if exit_code == 0:
        console.print("[bold green]:heavy_check_mark: License key found[/]")
    else:
        console.print("[bold yellow]:warning: No license key found. If you do not receive a license key, or if your existing license key has expired, please do not hesitate to reach out for support at hello@titanml.co.[/]")

        # ask the user to enter the license key
        license_key = Prompt.ask("[bold]Please enter your license key: [/]", default="")
        if license_key == "":
            console.print("[bold red]:x: License key is required to start the server.[/]")
            return

        if not check_license_key_format(license_key):
            console.print("[bold red]:x: Invalid license key format. Please enter a valid license key.[/]")
            return

        license_res = run_license_validator(license_key)
        exit_code = license_res["exit_code"]
        if exit_code == 0:
            console.print("[bold green]:heavy_check_mark: License key validated successfully.[/]")
        else:
            console.print("[bold red]:x: License key validation failed.[/] Please enter a valid license key.")
            return
    
    console.print(
        Panel(
            "Welcome to Takeoff CLI! :rocket:\n\n"
            "Takeoff is a tool to deploy and manage any open source LLM models in a Docker container.\n"
            "You can use this CLI to start and manage Takeoff servers.\n\n"
            "Let's get started! :point_right:",
            title="Takeoff CLI",
            border_style="green",
            expand=False,
        )
    )

    # ──────────────────────────────── ask for model name and device ───────────────────────────────── #
    console.print("[bold]⚙ Please select a model to use:[/bold]", style="bold")
    model_name = get_model_from_user(DEFAULT_MODELS)

    if model_name == "Others":
        model_name = Prompt.ask("[bold]⚙ Please enter the model name: [/]", default="", show_default=False)
        if model_name == "":
            console.print("[bold red]:x: Model name is required to start the server.[/]")
            return

    device = Prompt.ask("[bold]⚙ Please enter the device to use: [/]", choices=["cpu", "gpu"], default="gpu")

    if device == "gpu":
        if not has_gpu:
            console.print(
                "[bold yellow]:warning: You have chosen to use a GPU, but no Nvidia GPU is detected. We recommend using a GPU for optimal performance."
            )
            confirm = Prompt.ask(
                "[bold]Are you sure you want to continue with a CPU?[/] (y/n)", choices=["y", "n"], default="n"
            )
            if confirm == "n":
                return
            else:
                device = "cpu"
                tag = "latest-cpu"
                console.print(f"Starting Takeoff server with [bold blue]{model_name}[/] on [bold blue]CPU[/]...\n")
        else:
            device = "cuda"
            tag = "latest-gpu"

            console.print(f"Starting Takeoff server with [bold blue]{model_name}[/] on [bold blue]{gpu_name}[/]...\n")
    elif device == "cpu":
        tag = "latest-cpu"
        # warn the user about using cpu
        console.print(
            "[yellow]:warning: You have chosen to use a CPU. We will deploy the model on a CPU, but we [bold]strongly[/] recommend using a GPU for optimal performance."
        )
        console.print(f"Starting Takeoff server with [bold blue]{model_name}[/] on [bold blue]CPU[/]...\n")

    # ────────────────────────────────────── pull takeoff image ────────────────────────────────────── #
    image_name = "tytn/takeoff"
    images = client.images.list(name=f"{image_name}:{tag}")
    if images:
        console.print(f"[bold green]:heavy_check_mark:[/] Image found: [bold blue]{image_name}:{tag}[/] locally.\n")
        os.environ["TAKEOFF_IMAGE"] = f"{image_name}:{tag}"
    else:
        with Progress(
            SpinnerColumn(finished_text="[bold green]✔[/]"),
            TextColumn(f"Pulling latest takeoff container [bold blue]{image_name}:{tag}[/] from Docker Hub..."),
            transient=True,
        ) as progress:
            try:
                task1 = progress.add_task("Pulling image...", model=model_name, device=device)
                pull_image(image_name, tag)
                progress.update(task1, completed=True)
                progress.console.print("[bold green]:heavy_check_mark:[/] Image pulled successfully.")
                os.environ["TAKEOFF_IMAGE"] = f"{image_name}:{tag}"
            except Exception as e:
                progress.console.print(f"Image pull failed: {e}")
                return

    # ──────────────────────────────────── launch takeoff server ───────────────────────────────────── #
    with Progress(
        SpinnerColumn(finished_text="[bold green]✔[/]"),
        TextColumn(
            "Deploying [bold blue]{task.fields[model]}[/] on [bold blue]{task.fields[device]}[/] "
            "[yellow](This may take a few minutes depending on the model size and your internet connection.)[/]"
        ),
        transient=True,
    ) as progress:
        try:
            task1 = progress.add_task("Starting server...", model=model_name, device=device)
            takeoff = Takeoff(model_name=model_name, device=device, license_key=license_key)
            takeoff.start()
            progress.update(task1, completed=True)
            progress.console.print("[bold green]:heavy_check_mark:[/] Server deployed successfully.\n")
            progress.console.print("🚀[bold green]Takeoff Server started[/]🚀")
            progress.console.print(f"Takeoff Server running at: [bold blue]{takeoff.server_url}[/]")
            progress.console.print(f"Management server running at: [bold blue]{takeoff.management_url}[/]")
        except Exception as e:
            progress.console.print(f"Server start failed: {e}")
            return

    console.print(f"Now you can navigate to [bold blue]{takeoff.server_url}[/] to interact with the model.")


# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #
#                                                 Takeoff Run Command                                                  #
# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #


@main.command(help="Start a Takeoff server with options.")
def run(
    model_name: str = typer.Option(..., "--model", "-m", help="The models to optimize."),
    device: str = typer.Option(..., "--device", "-d", help="The device to use."),
    max_batch_size: int = typer.Option(None, "--max_batch_size", help="The maximum batch size."),
    max_seq_len: int = typer.Option(None, "--max_seq_len", help="The maximum sequence length."),
    tensor_parallel: int = typer.Option(None, "--tensor_parallel", help="The number of tensor parallelism to use"),
    access_token: str = typer.Option(None, "--hf_access_token", help="The HF Access token for gated models."),
    license_key: str = typer.Option(None, "--license_key", help="The license key."),
):
    try:
        takeoff = Takeoff(
            model_name=model_name,
            device=device,
            max_batch_size=max_batch_size,
            max_seq_len=max_seq_len,
            tensor_parallel=tensor_parallel,
            access_token=access_token,
            license_key=license_key,
        )
        takeoff.start()
    except Exception as e:
        typer.echo(f"Server start failed: {e}")


# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #
#                                                 Takeoff List Command                                                 #
# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #


@main.command(help="List all running takeoff server containers.")
def list():
    # ─────────────────────────────────── check docker environment ─────────────────────────────────── #
    docker_installed = check_docker_environment()
    if not docker_installed:
        console.print(
            "[bold red]:x: Docker environment check failed. Please make sure Docker is installed and running.[/]"
        )
        return
    console.print("[bold green]:heavy_check_mark: Docker environment check passed[/]")

    # Create a Docker client
    import docker

    client = docker.from_env()

    # List all running containers
    containers = client.containers.list()

    table = Table()
    table.add_column("ID", style="cyan")
    table.add_column("Container Name", style="magenta")

    # Print the list of running containers
    for container in containers:
        if "takeoff" in container.name:
            table.add_row(container.short_id, container.name)
    console.print(table)


# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #
#                                                 Takeoff Down Command                                                 #
# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #


@main.command(help="Shut down a running takeoff server container.")
def down(
    id: str = typer.Option(None, "--id", "-i", help="The ID of the server to shut down."),
    is_all: bool = typer.Option(False, "--all", "-a", help="Shut down all running servers."),
):
    # ─────────────────────────────────── check docker environment ─────────────────────────────────── #
    docker_installed = check_docker_environment()
    if not docker_installed:
        console.print(
            "[bold red]:x: Docker environment check failed. Please make sure Docker is installed and running.[/]"
        )
        return
    console.print("[bold green]:heavy_check_mark: Docker environment check passed[/]")

    # Create a Docker client
    import docker

    client = docker.from_env()
    if id is None and not is_all:
        console.print(
            Panel(
                "[bold red]:x: No container ID provided.[/]\n"
                "[bold yellow]Please provide a container ID or use the --all flag to shut down all running servers.[/]\n\n"
                "[Example]: `takeoff down --id <container_id>` or `takeoff down --all`\n"
                "[Note]: You can get the server ID by running `takeoff list`\n"
                "[Help]: Run `takeoff down --help` for more information.\n",
                title="Error",
                border_style="red",
                expand=False,
            )
        )
        return

    if id is not None:
        console.print(f"Shutting down server with ID {id}...")
        try:
            client.containers.get(id).stop()
            console.print("[bold green]:heavy_check_mark: Server shut down successfully.[/]")
        except Exception as e:
            console.print(f"[bold red]:x: Server shut down failed: {e}[/]")

    if is_all:
        typer.echo("Shutting down all servers...")
        # List all running containers
        containers = client.containers.list()

        # Print the list of running containers
        for container in containers:
            if "takeoff" in container.name:
                console.print(f"Shutting down server with ID {container.short_id}, Name: {container.name}")

                client.containers.get(container.name).stop()

        console.print("[bold green]:heavy_check_mark: All servers shut down successfully.[/]")


if __name__ == "__main__":
    main()



"""
This module contains the data model for the Takeoff SDK. This is used to parse the environment variables
"""

# ───────────────────────────────────────────────────── imports ────────────────────────────────────────────────────── #

from typing import Dict, Optional

from pydantic import BaseModel, ConfigDict, Field, validator

# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #
#                                               Configuration Data Model                                               #
# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #

__all__ = ["TakeoffEnvSetting"]


class TakeoffEnvSetting(BaseModel):
    """
    This class contains the data model for the Takeoff SDK. This setting is used to set the environment variables
    """

    model_config = ConfigDict(extra="forbid", protected_namespaces=())

    # ------------------------------------- model configuration -------------------------------------- #

    # NOTE: these are expected to be set by the user
    model_name: str = Field(..., env="TAKEOFF_MODEL_NAME")
    device: str = Field(..., env="TAKEOFF_DEVICE")

    backend: Optional[str] = Field(None, env="TAKEOFF_BACKEND")

    max_batch_size: Optional[int] = Field(None, env="TAKEOFF_MAX_BATCH_SIZE")  # 8
    batch_duration_millis: Optional[int] = Field(None, env="TAKEOFF_BATCH_DURATION_MILLIS")  # 100
    access_token: Optional[str] = Field(None, env="TAKEOFF_ACCESS_TOKEN")  # None
    cuda_visible_devices: Optional[str] = Field(None, env="TAKEOFF_CUDA_VISIBLE_DEVICES")  # 0

    refill_threshold: Optional[float] = Field(None, env="TAKEOFF_REILL_THRESHOLD")  # 0.5
    disable_continuous_generation: Optional[bool] = Field(None, env="TAKEOFF_DISABLE_CONTINUOUS_GENERATION")  # False
    consumer_group: Optional[str] = Field(None, env="TAKEOFF_CONSUMER_GROUP")  # primary

    redis_host: Optional[str] = Field(None, env="TAKEOFF_REDIS_HOST")  # localhost
    redis_port: Optional[int] = Field(None, env="TAKEOFF_REDIS_PORT")  # 6379

    heartbeat_wait_interval: Optional[int] = Field(None, env="TAKEOFF_HEARTBEAT_WAIT_INTERVAL")  # 10

    # JF Config
    nvlink_unavailable: Optional[int] = Field(None, env="TAKEOFF_NVLINK_UNAVAILABLE")  # 0

    tensor_parallel: Optional[int] = Field(None, env="TAKEOFF_TENSOR_PARALLEL")  # 1
    disable_cuda_graph: Optional[int] = Field(None, env="TAKEOFF_DISABLE_CUDA_GRAPH")  # 0
    quant_type: Optional[str] = Field(None, env="TAKEOFF_QUANT_TYPE")  # auto, awq
    max_seq_len: Optional[int] = Field(None, env="TAKEOFF_MAX_SEQUENCE_LENGTH")  # 0
    disable_static: Optional[int] = Field(None, env="TAKEOFF_DISABLE_STATIC")  # 0

    export_traces: Optional[bool] = Field(None, env="TAKEOFF_EXPORT_TRACES")  # False
    traces_host: Optional[str] = Field(None, env="TAKEOFF_TRACES_HOST")  # http://localhost:4317

    is_echo: Optional[bool] = Field(None, env="TAKEOFF_ECHO")  # False

    log_level: Optional[str] = Field(None, env="TAKEOFF_LOG_LEVEL")  # INFO
    rust_log_level: Optional[str] = Field(None, env="RUST_LOG")  # INFO

    run_name: Optional[str] = Field(None, env="TAKEOFF_RUN_NAME")  # takeoff_default

    license_key: Optional[str] = Field(None, env="LICENSE_KEY")  # None
    allow_remote_images: Optional[bool] = Field(None, env="TAKEOFF_ALLOW_REMOTE_IMAGES")  # False

    @validator("device")
    def validate_device(cls, v):
        if v not in ["cpu", "cuda", "api"]:
            raise ValueError("device must be one of 'cpu', 'cuda', or 'api'")
        return v

    def settings_to_env_vars(self) -> Dict[str, str]:
        env_var_mapping = {}
        for field_name, model_field in self.__dict__.items():
            field_info = self.model_fields[field_name]
            default_value = field_info.default
            # Only add to env_var_mapping if the current value is different from the default value
            if model_field != default_value:
                if field_info.json_schema_extra is not None and "env" in field_info.json_schema_extra:
                    env_var_name = field_info.json_schema_extra["env"]

                    # Check if the value is of type bool and convert to lowercase string
                    value = model_field
                    if isinstance(value, bool):
                        value = str(value).lower()

                    env_var_mapping[env_var_name] = value

        return env_var_mapping



import re


def extract_latest_version(changelog_path: str) -> str:
    """Extract the latest release version from the changelog file

    Args:
        changelog_path (str): path to the changelog file

    Returns:
        str: the latest release version
    """
    # it has to follow the pattern: ## [x.y.z] - yyyy-mm-dd
    version_pattern = re.compile(r"## \[(\d+\.\d+\.\d+)\] - \d{4}-\d{2}-\d{2}")
    with open(changelog_path, "r", encoding="utf-8") as file:
        for line in file:
            match = version_pattern.match(line)
            if match:
                return match.group(1)  # Return the first matched version
    return None  # Return None if no version is found


if __name__ == "__main__":
    changelog_path = "CHANGELOG.md"  # Update this path to your actual changelog file
    latest_version = extract_latest_version(changelog_path)
    if latest_version:
        print(f"Latest release version: {latest_version}")
    else:
        print("No release version found in the changelog.")
        raise SystemExit(1)

    with open("version.txt", "w") as version_file:
        version_file.write(latest_version)
    print(f"Version written to version.txt: {latest_version}")



import pytest
from pydantic import ValidationError

from takeoff_launcher import TakeoffEnvSetting


def test_data_model_with_default_values():
    # Test with default values
    settings = TakeoffEnvSetting(model_name="default_model", device="cpu")
    assert settings.model_name == "default_model"
    assert settings.device == "cpu"

    # assert everything else is None
    for field_name, model_field in settings.__dict__.items():
        if field_name not in ["model_name", "device"]:
            assert model_field is None


def test_settings_to_env_vars_with_custom_values():
    # Test with custom values
    settings = TakeoffEnvSetting(model_name="custom_model", device="cpu", max_batch_size=16)
    env_vars = settings.settings_to_env_vars()
    expected_vars = {"TAKEOFF_MODEL_NAME": "custom_model", "TAKEOFF_DEVICE": "cpu", "TAKEOFF_MAX_BATCH_SIZE": 16}
    assert env_vars == expected_vars


def test_settings_to_env_vars_with_bool_values():
    # Test with bool values, it shoud be converted to lower case string
    settings = TakeoffEnvSetting(model_name="custom_model", device="cpu", disable_continuous_generation=True)
    env_vars = settings.settings_to_env_vars()
    expected_vars = {
        "TAKEOFF_MODEL_NAME": "custom_model",
        "TAKEOFF_DEVICE": "cpu",
        "TAKEOFF_DISABLE_CONTINUOUS_GENERATION": "true",
    }
    assert env_vars == expected_vars


def test_validate_device():
    # Valid devices should not raise an error
    valid_devices = ["cpu", "cuda", "api"]
    for device in valid_devices:
        settings = TakeoffEnvSetting(model_name="test_model", device=device)
        assert settings.device == device

    # Invalid device should raise an error
    with pytest.raises(ValueError) as e:
        TakeoffEnvSetting(model_name="test_model", device="invalid_device")
    assert "device must be one of 'cpu', 'cuda', or 'api'" in str(e.value)


def test_invalid_attribute():
    # Attempting to create an instance with an undefined attribute should raise an error
    with pytest.raises(ValidationError):
        TakeoffEnvSetting(model_name="test_model", device="cpu", undefined_attribute="value")
